import { MaybeRef, Ref, computed, isRef, ref } from "vue";

export type BaseQueryKey = string | `${string}:${string | number | null}`;

type FilterKey = `filter[${string}]`;
type Filter = Record<FilterKey, Ref<any>>;
type FilterRaw = Record<FilterKey, any>;

export interface QueryParams<QueryKey extends BaseQueryKey> {
  size?: MaybeRef<number>;
  additionalKeys?: MaybeRef<QueryKey[]>;
  page?: MaybeRef<number>;
  append?: MaybeRef<string[]>;
  include?: MaybeRef<string[]>;
  sort?: MaybeRef<string>;
  filters?: MaybeRef<Record<string, MaybeRef<any>>>;
  body?: MaybeRef<Record<string, any>>;
}

export class QueryBuilder<QueryKey extends BaseQueryKey> {
  _keys: Ref<QueryKey[]>;
  _path: Ref<string>;
  _filters: Filter = {};
  _pageNumber: Ref<number> | null = null;
  _pageSize: Ref<number> | null = null;
  _sort: Ref<string> | null = null;
  _append: Ref<string[]> = ref([]);
  _include: Ref<string[]> = ref([]);
  _body: Ref<Record<string, any>> = ref({});

  constructor(queryKeys: MaybeRef<QueryKey[]>, path: MaybeRef<string>) {
    if (isRef(queryKeys)) {
      this._keys = queryKeys;
    } else {
      this._keys = ref(queryKeys) as Ref<QueryKey[]>;
    }

    if (isRef(path)) {
      this._path = path;
    } else {
      this._path = ref(path);
    }
  }

  get hasPageNumber(): boolean {
    return this._pageNumber != null;
  }

  get hasPageSize(): boolean {
    return this._pageSize != null;
  }

  get hasSort(): boolean {
    return this._sort != null;
  }

  get hasFilters(): boolean {
    return Object.keys(this._filters).length > 0;
  }

  get hasAppend(): boolean {
    return this._append.value.length > 0;
  }

  get hasInclude(): boolean {
    return this._include.value.length > 0;
  }

  get hasBody(): boolean {
    return !!this._body;
  }

  get params(): Ref<{
    include?: string[] | undefined;
    append?: string[] | undefined;
    filter?: FilterRaw | undefined;
    sort?: string | null | undefined;
    "page[number]"?: number | null | undefined;
    "page[size]"?: number | null | undefined;
  }> {
    return computed(() => {
      const params: {
        [key: FilterKey]: number | string | boolean | null;
        include?: string[];
        append?: string[];
        sort?: string | null;
        "page[number]"?: number | null;
        "page[size]"?: number | null;
      } = {};

      if (this.hasFilters) {
        Object.entries(this._filters).forEach(([key, value]) => {
          params[key as FilterKey] = value.value;
        });
      }

      if (this.hasPageSize) {
        params["page[size]"] = this._pageSize?.value;
      }

      if (this.hasPageNumber) {
        params["page[number]"] = this._pageNumber?.value;
      }

      if (this.hasSort) {
        params.sort = this._sort?.value;
      }

      if (this.hasAppend) {
        params.append = this._append.value;
      }

      if (this.hasInclude) {
        params.include = this._include.value;
      }

      return params;
    });
  }

  get path(): Ref<string> {
    return this._path;
  }

  get body(): Ref<Record<string, any>> {
    return this._body;
  }

  set body(body: MaybeRef<Record<string, any>> | undefined) {
    if (body === undefined) return;

    if (isRef<Record<string, any>>(body)) {
      this._body = body;
    } else {
      this._body = ref(body);
    }
  }

  get keys(): Ref<(QueryKey | [string, any])[]> {
    return computed(() => {
      const mainQueryKeys: QueryKey[] = this._keys.value;

      // If any of the query keys contain an object, this is an error and should be caught
      // Catch this by looking for `[object Object]` in a key string
      if (
        mainQueryKeys.some((key) => key.toString().includes("[object Object]"))
      ) {
        throw new Error("Query keys should not contain objects, only strings");
      }

      let keys: (QueryKey | [string, any])[] = [...mainQueryKeys];

      if (this.hasFilters) {
        const filters = Object.entries(this._filters).map<[string, any]>(
          ([key, value]) => [key, value.value]
        );

        keys = [...keys, ...filters];
      }

      if (this.hasPageSize) {
        keys = [...keys, ["page[size]", this._pageSize?.value]];
      }

      if (this.hasPageNumber) {
        keys = [...keys, ["page[number]", this._pageNumber?.value]];
      }

      if (this.hasSort) {
        keys = [...keys, ["sort", this._sort?.value]];
      }

      if (this.hasAppend) {
        keys = [...keys, ["append", this._append.value]];
      }

      if (this.hasInclude) {
        keys = [...keys, ["include", this._include.value]];
      }

      return keys;
    });
  }

  static keys<QueryKey extends BaseQueryKey>(queryKey: QueryKey[]): QueryKey[] {
    return queryKey;
  }

  static key<QueryKey extends BaseQueryKey>(queryKey: QueryKey): QueryKey {
    return queryKey;
  }

  filter(
    key: string,
    value?: MaybeRef<string | number | null | boolean | undefined>
  ) {
    if (value === undefined) return;
    if (value === null) return;

    if (isRef(value)) {
      this._filters[`filter[${key}]`] = value;
      return this;
    } else {
      this._filters[`filter[${key}]`] = ref(value);
      return this;
    }
  }

  filterIf(
    condition: boolean,
    key: string,
    value?: MaybeRef<string | number | null | boolean>
  ) {
    if (!condition) return;
    this.filter(key, value);
  }

  sort(sort?: MaybeRef<string>) {
    if (sort === undefined) return;
    if (isRef(sort)) this._sort = sort;
    else this._sort = ref(sort);
  }

  page(page?: MaybeRef<number>) {
    if (page === undefined) return;
    if (isRef(page)) this._pageNumber = page;
    else this._pageNumber = ref(page);
  }

  size(size?: MaybeRef<number>) {
    if (size === undefined) return;
    if (isRef(size)) this._pageSize = size;
    else this._pageSize = ref(size);
  }

  withKey(key?: MaybeRef<QueryKey>) {
    if (key === undefined) return;
    this._keys.value.push(isRef(key) ? key.value : key);
  }

  withKeys(key?: MaybeRef<QueryKey[]>) {
    if (key === undefined) return;
    this._keys.value.push(...(isRef(key) ? key.value : key));
  }

  withKeyIf(condition: boolean, key: MaybeRef<QueryKey>) {
    if (!condition) return;
    this.withKey(key);
  }

  withKeysIf(condition: boolean, key: MaybeRef<QueryKey[]>) {
    if (!condition) return;
    this.withKeys(key);
  }

  append(append?: MaybeRef<string> | MaybeRef<string[]>) {
    if (append === undefined) return;

    if (isRef(append)) {
      if (Array.isArray(append.value)) {
        this._append.value.push(...append.value);
      } else {
        this._append.value.push(append.value);
      }
    } else {
      if (Array.isArray(append)) {
        this._append.value.push(...append);
      } else {
        this._append.value.push(append);
      }
    }
  }

  include(include?: MaybeRef<string> | MaybeRef<string[]>) {
    if (include === undefined) return;

    if (isRef(include)) {
      if (Array.isArray(include.value)) {
        this._include.value.push(...include.value);
      } else {
        this._include.value.push(include.value);
      }
    } else {
      if (Array.isArray(include)) {
        this._include.value.push(...include);
      } else {
        this._include.value.push(include);
      }
    }
  }

  withBody(body: MaybeRef<Record<string, any>>) {
    if (isRef<Record<string, any>>(body)) {
      this._body = body;
    } else {
      this._body = ref(body);
    }
  }

  /**
   * Alias for withBody
   */
  withPayload(payload: MaybeRef<Record<string, any>>) {
    this.withBody(payload);
  }

  /**
   * Build the query params for the query.
   */
  build(params: QueryParams<QueryKey>) {
    this.size(params.size);
    this.withKeys(params.additionalKeys);
    this.page(params.page);
    this.append(params.append);
    this.include(params.include);
    this.sort(params.sort);

    if (params.body) this.withBody(params.body);

    Object.entries(
      (isRef<Record<string, MaybeRef<any>>>(params.filters)
        ? params.filters.value
        : params.filters) || {}
    ).forEach(([key, value]) => {
      this.filter(key, value);
    });
  }

  static mutate<QueryKey extends BaseQueryKey>(
    path: string
  ): QueryBuilder<QueryKey> {
    return new QueryBuilder([], path);
  }
}
